import ast


def get_auxiliary_name(node: ast.AST, aux_symbol_id: str) -> str:
    """
    Generates a name for auxiliary variables.

    :param node: the ast node that originates the auxiliary symbol
    :param aux_symbol_id: the id name of the auxiliary symbol
    :return: the unique name to the symbol.
    """
    return "%s_%s" % (aux_symbol_id, id(node))



from abc import ABC
from typing import Optional


class CompilerWarning(ABC):
    def __init__(self, line: int, col: int):
        self.line: int = line
        self.col: int = col

    @property
    def message(self) -> str:
        message = '' if self._warning_message is None else ' - ' + self._warning_message
        return '{0}:{1}{2}'.format(self.line, self.col, message)

    @property
    def _warning_message(self) -> Optional[str]:
        return None

    def __str__(self) -> str:
        return self.message


class NameShadowing(CompilerWarning):
    """
    An warning raised when a name from an outer scope symbol is used as the name of an inner scope symbol
    """
    from boa3.model.symbol import ISymbol

    def __init__(self, line: int, col: int, outer_symbol: ISymbol, symbol_id: str):
        self.symbol_id: str = symbol_id
        self.existing_symbol = outer_symbol
        super().__init__(line, col)

    @property
    def _warning_message(self) -> Optional[str]:
        if self.symbol_id is not None:
            return "Shadowing {0} name '{1}'".format(self.existing_symbol.shadowing_name, self.symbol_id)



from typing import List, Tuple

from boa3.model.operation.operator import Operator
from boa3.model.operation.unary.unaryoperation import UnaryOperation
from boa3.model.type.type import IType, Type
from boa3.neo.vm.opcode.Opcode import Opcode


class NoneNotIdentity(UnaryOperation):
    """
    A class used to represent an 'is not None' comparison

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar operand: the operand type. Inherited from :class:`UnaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """

    def __init__(self, operand: IType = Type.int):
        self.operator: Operator = Operator.IsNot
        super().__init__(operand)

    @property
    def number_of_operands(self) -> int:
        # it is a Python binary operation that is equivalent to a Neo VM unary operation
        return 2

    @property
    def op_on_stack(self) -> int:
        return super().number_of_operands

    def validate_type(self, *types: IType) -> bool:
        if len(types) == self.op_on_stack:
            return True
        if len(types) != self.number_of_operands:
            return False
        left: IType = types[0]
        right: IType = types[1]

        return left is Type.none or right is Type.none

    def _get_result(self, operand: IType) -> IType:
        if self.validate_type(operand):
            return Type.bool
        else:
            return Type.none

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        return [
            (Opcode.ISNULL, b''),
            (Opcode.NOT, b'')
        ]



import argparse
import logging
import os
import sys

from boa3.boa3 import Boa3


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input", help=".py smart contract to compile")
    args = parser.parse_args()

    if not args.input.endswith(".py") or not os.path.isfile(args.input):
        logging.error("Input file is not .py")
        sys.exit(1)

    fullpath = os.path.realpath(args.input)
    path, filename = os.path.split(fullpath)

    try:
        Boa3.compile_and_save(args.input)
        logging.info(f"Wrote {filename.replace('.py', '.nef')} to {path}")
    except Exception as e:
        logging.exception(e)


if __name__ == "__main__":
    main()



from typing import Dict

from boa3.model.expression import IExpression
from boa3.model.method import Method
from boa3.model.symbol import ISymbol
from boa3.model.variable import Variable


class Module(ISymbol):
    """
    A class used to represent a Python module

    :ivar variables: a dictionary that maps each variable with its name. Empty by default.
    :ivar methods: a dictionary that maps each method with its name. Empty by default.
    """

    def __init__(self, variables: Dict[str, Variable] = None, methods: Dict[str, Method] = None):
        if variables is None:
            variables = {}
        self.variables = variables

        if methods is None:
            methods = {}
        self.methods = methods

    @property
    def shadowing_name(self) -> str:
        return 'module'

    def include_variable(self, var_id: str, var: Variable):
        """
        Includes a variable into the scope of the module

        :param var_id: variable identifier
        :param var: variable to be included
        """
        if var_id not in self.symbols:
            self.variables[var_id] = var

    def include_method(self, method_id: str, method: Method):
        """
        Includes a method into the scope of the module

        :param method_id: method identifier
        :param method: method to be included
        """
        if method_id not in self.symbols:
            self.methods[method_id] = method

    @property
    def symbols(self) -> Dict[str, IExpression]:
        """
        Gets all the symbols in the module

        :return: a dictionary that maps each symbol in the module with its name
        """
        symbols = {}
        symbols.update(self.variables)
        symbols.update(self.methods)
        return symbols



def to_script_hash(data_bytes: bytes) -> bytes:
    """
    Converts a data to a script hash.

    :param data_bytes: data to hash.
    :type data_bytes: bytearray or bytes

    :return: the scripthash of the data
    :rtype: bytes
    """
    from boa3.neo import cryptography
    return cryptography.hash160(data_bytes)


def to_hex_str(data_bytes: bytes) -> str:
    """
    Converts bytes into its string hex representation.

    :param data_bytes: data to represent as hex.
    :type data_bytes: bytearray or bytes

    :return: the hex representation of the data
    :rtype: str
    """
    if isinstance(data_bytes, bytes):
        data_bytes = bytearray(data_bytes)
    data_bytes.reverse()
    return data_bytes.hex()



import ast
from typing import Dict

from boa3.analyser.constructanalyser import ConstructAnalyser
from boa3.analyser.moduleanalyser import ModuleAnalyser
from boa3.analyser.typeanalyser import TypeAnalyser
from boa3.model.symbol import ISymbol
from boa3.model.type.type import Type


class Analyser(object):
    """
    This class is responsible for the semantic analysis of the code

    :ivar symbol_table: a dictionary used to store the identifiers
    """

    def __init__(self, ast_tree: ast.AST):
        self.symbol_table: Dict[str, ISymbol] = {}

        self.ast_tree: ast.AST = ast_tree
        self.is_analysed: bool = False

        self.__include_builtins_symbols()

    @staticmethod
    def analyse(path: str):
        """
        Analyses the syntax of the Python code

        :param path: the path of the Python file
        :return: a boolean value that represents if the analysis was successful
        :rtype: Analyser
        """
        with open(path, 'rb') as source:
            ast_tree = ast.parse(source.read())

        analyser = Analyser(ast_tree)
        analyser.__pre_execute()
        # fill symbol table
        if not analyser.__analyse_modules():
            return analyser
        # check is the types are correct
        if not analyser.__check_types():
            return analyser
        analyser.is_analysed = True  # TODO

        return analyser

    def __include_builtins_symbols(self):
        """
        Include the Python builtins in the global symbol table
        """
        self.symbol_table.update(Type.values())

    def __check_types(self) -> bool:
        """
        Performs the type checking

        :return: a boolean value that represents if the analysis was successful
        """
        type_analyser = TypeAnalyser(self.ast_tree, self.symbol_table)
        return not type_analyser.has_errors

    def __analyse_modules(self) -> bool:
        """
        Validates the symbols and constructs the symbol table of the ast tree

        :return: a boolean value that represents if the analysis was successful
        """
        module_analyser = ModuleAnalyser(self.ast_tree, self.symbol_table)
        self.symbol_table.update(module_analyser.global_symbols)
        return not module_analyser.has_errors

    def __pre_execute(self):
        """
        Pre executes the instructions of the ast for optimization
        """
        self.ast_tree = ConstructAnalyser(self.ast_tree).tree



from abc import ABC, abstractmethod
from typing import Dict

from boa3.model.expression import IExpression
from boa3.model.method import Method
from boa3.model.type.itype import IType
from boa3.model.variable import Variable


class IBuiltinDecorator(Method, ABC):
    def __init__(self, identifier: str, args: Dict[str, Variable] = None, return_type: IType = None):
        self.identifier = identifier
        super().__init__(args, return_type)

    @abstractmethod
    def validate_parameters(self, *params: IExpression) -> bool:
        """
        Verifies if the given parameters are valid to the method

        :param params: arguments of the method
        :return: True if all arguments are valid. False otherwise.
        """
        pass



from typing import List, Tuple

from boa3.model.operation.binary.binaryoperation import BinaryOperation
from boa3.model.operation.operator import Operator
from boa3.model.type.type import IType, Type
from boa3.neo.vm.opcode.Opcode import Opcode


class GreaterThanOrEqual(BinaryOperation):
    """
    A class used to represent a numeric greater than or equal comparison

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar left: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar right: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: List[IType] = [Type.int, Type.str]

    def __init__(self, left: IType = Type.int, right: IType = None):
        self.operator: Operator = Operator.GtE
        super().__init__(left, right)

    def validate_type(self, *types: IType) -> bool:
        if len(types) != self.number_of_operands:
            return False
        left: IType = types[0]
        right: IType = types[1]

        # TODO: change the logic of the validation when implement other numeric types
        return left == right and left in self._valid_types

    def _get_result(self, left: IType, right: IType) -> IType:
        if self.validate_type(left, right):
            return Type.bool
        else:
            return Type.none

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        return [(Opcode.GE, b'')]



from typing import Dict, Optional

from boa3.model.expression import IExpression
from boa3.model.type.itype import IType
from boa3.model.variable import Variable


class Method(IExpression):
    """
    A class used to represent a function or a class method

    :ivar args: a dictionary that maps each arg with its name. Empty by default.
    :ivar locals: a dictionary that maps each local variable with its name. Empty by default.
    :ivar is_public: a boolean value that specifies if the method is public. False by default.
    :ivar return_type: the return type of the method. None by default.
    """

    def __init__(self, args: Dict[str, Variable] = None, return_type: IType = None, is_public: bool = False):
        from boa3.neo.vm.VMCode import VMCode
        if args is None:
            args = {}
        self.args: Dict[str, Variable] = args
        self.return_type: IType = return_type

        self.is_public: bool = is_public
        self.is_main_method: bool = False

        self.locals: Dict[str, Variable] = {}
        self.init_bytecode: Optional[VMCode] = None

    @property
    def shadowing_name(self) -> str:
        return 'method'

    def include_variable(self, var_id: str, var: Variable):
        """
        Includes a variable into the list of locals

        :param var_id: variable identifier
        :param var: variable to be included
        """
        if var_id not in self.symbols:
            self.locals[var_id] = var

    @property
    def type(self) -> IType:
        return self.return_type

    @property
    def symbols(self) -> Dict[str, Variable]:
        """
        Gets all the symbols in the module

        :return: a dictionary that maps each symbol in the module with its name
        """
        symbols = {}
        symbols.update(self.args)
        symbols.update(self.locals)
        return symbols

    @property
    def bytecode_address(self) -> Optional[int]:
        """
        Gets the address where this method starts in the bytecode

        :return: the first address of the method
        """
        if self.init_bytecode is None:
            return None
        else:
            return self.init_bytecode.start_address

    def set_as_main_method(self):
        self.is_main_method = True
        self.is_public = True



from abc import ABC
from typing import Iterable, Optional, Union


class CompilerError(ABC, Exception):
    """
    An interface for compilation errors
    """

    def __init__(self, line: int, col: int):
        self.line: int = line
        self.col: int = col

    @property
    def message(self) -> str:
        message = '' if self._error_message is None else ' - ' + self._error_message
        return '{0}:{1}{2}'.format(self.line, self.col, message)

    @property
    def _error_message(self) -> Optional[str]:
        return None

    def __str__(self) -> str:
        return self.message


class TypeHintMissing(CompilerError):
    """
    An error raised when type hint cannot be found
    """

    def __init__(self, line: int, col: int, symbol_id: str = None):
        self.symbol_id = symbol_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        if self.symbol_id is not None:
            return "Type hint is missing for the symbol '%s'" % self.symbol_id


class InvalidType(CompilerError):
    """
    An error raised when a type that is not supported by Neo VM is used
    """

    def __init__(self, line: int, col: int, symbol_id: str = None):
        self.symbol_id = symbol_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        message = "Invalid type"
        if self.symbol_id is not None:
            message += ": '%s'" % self.symbol_id
        return message


class NotSupportedOperation(CompilerError):
    """
    An error raised when an operation that is not supported by Neo VM is used
    """

    def __init__(self, line: int, col: int, symbol_id: str):
        self.symbol_id = symbol_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "The following operation is not supported: '%s'" % self.symbol_id


class UnresolvedReference(CompilerError):
    """
    An error raised when an undefined symbol is used
    """

    def __init__(self, line: int, col: int, symbol_id: str):
        self.symbol_id = symbol_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Unresolved reference '%s'" % self.symbol_id


class UnresolvedOperation(CompilerError):
    """
    An error raised when an undefined symbol is used
    """

    def __init__(self, line: int, col: int, type_id: str, operation_id: str):
        self.type_id = type_id
        self.operation_id = operation_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Unresolved reference: '%s' does not have a definition of '%s' operator" % (self.type_id, self.operation_id)


class MismatchedTypes(CompilerError):
    """
    An error raised when the evaluated and expected types are not the same
    """

    def __init__(self, line: int, col: int, expected_type_id: Union[str, Iterable[str]], actual_type_id: Union[str, Iterable[str]]):
        if isinstance(expected_type_id, str):
            expected_type_id = [expected_type_id]
        if isinstance(actual_type_id, str):
            actual_type_id = [actual_type_id]

        self.expected_types = expected_type_id
        self.actual_types = actual_type_id
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        expected_types = join_args(self.expected_types)
        actual_types = join_args(self.actual_types)
        return "Expected type '%s', got '%s' instead" % (expected_types, actual_types)


class TooManyReturns(CompilerError):
    """
    An error raised when a function returns a tuple
    """

    def __init__(self, line: int, col: int):
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Too many returns"


class IncorrectNumberOfOperands(CompilerError):
    """
    An error raised when an operation is used with the wrong number of operands
    """

    def __init__(self, line: int, col: int, expected_count: int, actual_count: int):
        self.expected = expected_count
        self.actual = actual_count
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Incorrect number of operands: expected '%s', got '%s' instead" % (self.expected, self.actual)


class UnexpectedArgument(CompilerError):
    """
    An error thrown when more arguments are used in a function than the number of arguments in the function's signature
    """

    def __init__(self, line: int, col: int):
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Unexpected argument"


class UnfilledArgument(CompilerError):
    """
    An error thrown when less arguments are used in a function than the number of arguments in the function's signature
    """

    def __init__(self, line: int, col: int, param: str):
        self.param = param
        super().__init__(line, col)

    @property
    def _error_message(self) -> Optional[str]:
        return "Parameter '%s' unfilled" % self.param


def join_args(iterable: Iterable[str]) -> str:
    return str.join("', '", iterable)



import json
import logging
import sys
from typing import Any, Dict, List, Optional, Tuple

from boa3.model.method import Method
from boa3.model.symbol import ISymbol
from boa3.neo import to_hex_str
from boa3.neo.smart_contract.neffile import NefFile


class FileGenerator:
    """
    This class is responsible for generating the files.
    """

    def __init__(self, bytecode: bytes, symbols: Dict[str, ISymbol] = None):
        if symbols is None:
            symbols = {}
        self.__symbols: Dict[str, ISymbol] = symbols
        self.__nef: NefFile = NefFile(bytecode)

    @property
    def __nef_hash(self) -> str:
        """
        Gets the string representation of the hash of the nef file

        :return: the hex string representation of the hash
        """
        return '0x' + to_hex_str(self.__nef.script_hash)

    @property
    def __methods(self) -> Dict[str, Method]:
        """
        Gets a sublist of the symbols containing all the methods

        :return: a dictionary that maps each method with its identifier
        """
        return {name: method for name, method in self.__symbols.items() if isinstance(method, Method) and method.is_public}

    @property
    def __entry_point(self) -> Optional[Tuple[str, Method]]:
        """
        Gets the entry point method of the smart contract

        :return: a tuple with the name and the method if found. Otherwise, return None
        :rtype: Tuple[str, Method] or None
        """
        method_id = None
        if 'main' in self.__methods:
            method_id = 'main'
        elif 'Main' in self.__methods:
            method_id = 'Main'

        if method_id is None:
            raise NotImplementedError
        else:
            return method_id, self.__methods[method_id]

    def generate_nef_file(self) -> bytes:
        """
        Generates the .nef file

        :return: the resulting nef file as a byte array
        """
        return self.__nef.serialize()

    def generate_manifest_file(self) -> bytes:
        """
        Generates the .manifest metadata file

        :return: the resulting manifest as a byte array
        """
        data: Dict[str, Any] = self.__get_manifest_info()
        json_data: str = json.dumps(data, indent=4)
        return bytes(json_data, sys.getdefaultencoding())

    def __get_manifest_info(self) -> Dict[str, Any]:
        """
        Gets the manifest information in a dictionary format

        :return: a dictionary with the manifest information
        """
        # TODO: fill the information of the manifest
        return {
            "groups": [],
            "features": {
                "storage": False,
                "payable": False
            },
            "abi": self.__get_abi_info(),
            "permissions": [
                {
                    "contract": "*",
                    "methods": "*"
                }
            ],
            "trusts": [],
            "safeMethods": [],
            "extra": None
        }

    def generate_abi_file(self) -> bytes:
        """
        Generates the .abi metadata file

        :return: the resulting abi as a byte array
        """
        data: Dict[str, Any] = self.__get_abi_info()
        json_data: str = json.dumps(data)
        return bytes(json_data, sys.getdefaultencoding())

    def __get_abi_info(self) -> Dict[str, Any]:
        """
        Gets the abi information in a dictionary format

        :return: a dictionary with the abi information
        """
        return {
            "hash": self.__nef_hash,
            "methods": self.__get_abi_methods(),
            "entryPoint": self.__get_abi_entry_point(),
            "events": self.__get_abi_events()
        }

    def __get_abi_methods(self) -> List[Dict[str, Any]]:
        """
        Gets the abi methods in a dictionary format

        :return: a dictionary with the abi methods
        """
        entry_point = self.__entry_point
        if entry_point is not None:
            entry_point = entry_point[0]    # method id

        methods = []
        for method_id, method in self.__methods.items():
            if method_id != entry_point:
                logging.info("'{0}' method included in the ABI".format(method_id))
                methods.append(self.__construct_abi_method(method_id, method))
        return methods

    def __get_abi_entry_point(self) -> Dict[str, Any]:
        """
        Gets the abi entry point method in a dictionary format

        :return: a dictionary with the abi entry point
        """
        if self.__entry_point is None:
            return {}
        else:
            method_id, method = self.__entry_point
            return self.__construct_abi_method(method_id, method)

    def __construct_abi_method(self, method_id: str, method: Method) -> Dict[str, Any]:
        params = []
        for arg_id, arg in method.args.items():
            params.append({
                "name": arg_id,
                "type": arg.type.abi_type
            })
        return {
            "name": method_id,
            "parameters": params,
            "returnType": method.type.abi_type
        }

    def __get_abi_events(self) -> List[Dict[str, Any]]:
        """
        Gets the abi events in a dictionary format

        :return: a dictionary with the abi events
        """
        # TODO: abi events
        return []

    def generate_avmdbgnfo_file(self) -> bytes:
        """
        Generates a debug map for NEO debugger

        :return: the resulting map as a byte array
        """
        raise NotImplementedError



from typing import Any

from boa3.model.type.itype import IType
from boa3.neo.vm.type.AbiType import AbiType


class __AnyType(IType):
    """
    A class used to represent Python bool type
    """

    def __init__(self):
        identifier = 'any'
        super().__init__(identifier)

    @property
    def abi_type(self) -> AbiType:
        return AbiType.Any

    @classmethod
    def build(cls, value: Any):
        from boa3.model.type.type import Type
        return Type.any

    @classmethod
    def _is_type_of(cls, value: Any):
        return True


anyType: IType = __AnyType()



import ast
import logging
from abc import ABC
from typing import Any, Dict, List, Optional

from boa3.exception.CompilerError import CompilerError
from boa3.exception.CompilerWarning import CompilerWarning
from boa3.model.expression import IExpression
from boa3.model.operation.operation import IOperation
from boa3.model.symbol import ISymbol
from boa3.model.type.type import IType, Type


class IAstAnalyser(ABC, ast.NodeVisitor):
    """
    An interface for the analysers that walk the Python abstract syntax tree

    :ivar errors: a list that contains all the errors raised by the compiler. Empty by default.
    :ivar warnings: a list that contains all the warnings found by the compiler. Empty by default.
    """

    def __init__(self, ast_tree: ast.AST):
        self.errors: List[CompilerError] = []
        self.warnings: List[CompilerWarning] = []
        self._tree: ast.AST = ast_tree
        self.symbols: Dict[str, ISymbol] = {}

    @property
    def has_errors(self) -> bool:
        return len(self.errors) > 0

    def _log_error(self, error: CompilerError):
        if not any(err.message == error.message for err in self.errors):
            # don't include duplicated errors
            self.errors.append(error)
            logging.error(error)

    def _log_warning(self, warning: CompilerWarning):
        if not any(warn.message == warning.message for warn in self.errors):
            # don't include duplicated warnings
            self.warnings.append(warning)
            logging.warning(warning)

    def get_type(self, value: Any) -> IType:
        """
        Returns the type of the given value.

        :param value: value to get the type
        :return: Returns the :class:`IType` of the the type of the value. `Type.none` by default.
        """
        # visits if it is a node
        if isinstance(value, ast.AST):
            fun_rtype_id: str = ast.NodeVisitor.visit(self, value)
            if isinstance(fun_rtype_id, ast.Name):
                fun_rtype_id = fun_rtype_id.id

            if isinstance(fun_rtype_id, str) and not isinstance(value, ast.Str):
                value = self.get_symbol(fun_rtype_id)
            else:
                value = fun_rtype_id

        if isinstance(value, IType):
            return value
        elif isinstance(value, IExpression):
            return value.type
        elif isinstance(value, IOperation):
            return value.result
        else:
            return Type.get_type(value)

    def get_symbol(self, symbol_id: str) -> Optional[ISymbol]:
        """
        Tries to get the symbol by its id name

        :param symbol_id: the id name of the symbol
        :return: the symbol if found. None otherwise.
        :rtype: ISymbol or None
        """
        if symbol_id in self.symbols:
            # the symbol exists in the global scope
            return self.symbols[symbol_id]
        else:
            # the symbol may be a built in. If not, returns None
            from boa3.model.builtin.builtin import Builtin
            return Builtin.get_symbol(symbol_id)



from enum import Enum


class StackItemType(bytes, Enum):
    Any = b'\x00'
    Pointer = b'\x10'
    Boolean = b'\x20'
    Integer = b'\x21'
    ByteString = b'\x28'
    Buffer = b'\x30'
    Array = b'\x40'
    Struct = b'\x41'
    Map = b'\x48'
    InteropInterface = b'\x60'



from typing import Optional

from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.opcode.OpcodeInformation import OpcodeInformation


class VMCode:
    """
    Represents a Neo VM code

    :ivar opcode: the opcode of the code
    :ivar data: the data in bytes of the code. Empty byte array by default.
    """

    def __init__(self, op_info: OpcodeInformation, last_code=None, data: bytes = bytes()):
        """
        :param op_info: information of the opcode of the code
        :param last_code: the previous code in the bytecode. None if it's the first.
        :type last_code: VMCode or None
        :param data: the data in bytes of the code. Empty byte array by default.
        """
        self._last_code = last_code    # type:Optional[VMCode]
        self.info: OpcodeInformation = op_info
        self.data: bytes = self.__format_data(data)

    @property
    def start_address(self) -> int:
        """
        Gets the start address of this code

        :return: the first address of the code
        """
        if self._last_code is None:
            return 0
        else:
            return self._last_code.end_address + 1

    @property
    def end_address(self) -> int:
        """
        Gets the end address of this code

        :return: the last address of the code
        """
        return self.start_address + len(self.data)

    @property
    def opcode(self) -> Opcode:
        """
        Gets the Neo VM opcode of the code

        :return: the opcode of the code
        """
        return self.info.opcode

    def update(self, opcode: OpcodeInformation, data: bytes = bytes()):
        self.info = opcode
        self.data = self.__format_data(data)

    def __format_data(self, data: bytes) -> bytes:
        data_len: int = len(data)
        min_data_len: int = self.info.data_len
        max_data_len: int = self.info.max_data_len

        if data_len < min_data_len:
            data = bytes(bytearray(data).zfill(min_data_len))
        elif data_len > max_data_len:
            data = data_len[0:max_data_len]

        return data

    def __str__(self) -> str:
        return self.opcode.name + ' ' + self.data.hex()



import sys
from typing import Any, Dict, List, Optional, Tuple

from boa3.analyser.analyser import Analyser
from boa3.constants import ONE_BYTE_MAX_VALUE, TWO_BYTES_MAX_VALUE
from boa3.model.builtin.builtin import Builtin
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.method import Method
from boa3.model.operation.binaryop import BinaryOp
from boa3.model.operation.operation import IOperation
from boa3.model.symbol import ISymbol
from boa3.model.type.type import IType, Type
from boa3.model.variable import Variable
from boa3.neo.vm.VMCode import VMCode
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.opcode.OpcodeInfo import OpcodeInfo
from boa3.neo.vm.opcode.OpcodeInformation import OpcodeInformation
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.StackItemType import StackItemType


class CodeGenerator:
    """
    This class is responsible for generating the Neo VM bytecode

    :ivar symbol_table: a dictionary that maps the global symbols.
    """

    @staticmethod
    def generate_code(analyser: Analyser) -> bytes:
        """
        Generates the Neo VM bytecode using of the analysed Python code

        :param analyser: semantic analyser it tge Python code
        :return: the Neo VM bytecode
        """
        from boa3.compiler.codegeneratorvisitor import VisitorCodeGenerator

        generator = CodeGenerator(analyser.symbol_table)
        visitor = VisitorCodeGenerator(generator)
        visitor.visit(analyser.ast_tree)
        return generator.bytecode

    def __init__(self, symbol_table: Dict[str, ISymbol]):
        self.symbol_table: Dict[str, ISymbol] = symbol_table
        self.__vm_codes_without_order: List[VMCode] = []

        self.__current_method: Method = None
        self.__function_calls: List[Tuple[VMCode, Method]] = []  # calls of functions that haven't been converted yet
        self.__is_entry_point_converted: bool = False

        self._stack: List[IType] = []  # simulates neo execution stack

    @property
    def bytecode(self) -> bytes:
        """
        Gets the bytecode of the translated code

        :return: the generated bytecode
        """
        bytecode = bytearray()
        for code in self.__vm_codes:
            bytecode += code.opcode
            if code.data is not None:
                bytecode += code.data
        return bytes(bytecode)

    @property
    def __vm_codes(self) -> List[VMCode]:
        """
        Gets an ordered list of the vm codes

        :return: a list of vm codes ordered by the code address
        """
        return sorted(self.__vm_codes_without_order, key=lambda code: code.start_address)

    @property
    def __vm_codes_map(self) -> Dict[int, VMCode]:
        """
        Gets a map of the vm codes

        :return: a dictionary that maps each code to its start address
        """
        code_dict = {}
        for code in self.__vm_codes:
            code_dict[code.start_address] = code
        return code_dict

    @property
    def last_code(self) -> Optional[VMCode]:
        """
        Gets the last code in the bytecode

        :return: the last code. If the bytecode is empty, returns None
        :rtype: VMCode or None
        """
        if len(self.__vm_codes) > 0:
            return self.__vm_codes[-1]
        else:
            return None

    @property
    def address(self) -> int:
        if self.last_code is not None:
            return self.last_code.end_address + 1
        else:
            return 0

    @property
    def __args(self) -> List[str]:
        """
        Gets a list with the arguments names of the current method

        :return: A list with the arguments names
        """
        if self.__current_method is not None:
            return list(self.__current_method.args.keys())
        else:
            return []

    @property
    def __locals(self) -> List[str]:
        """
        Gets a list with the variables names in the scope of the current method

        :return: A list with the variables names
        """
        if self.__current_method is not None:
            return list(self.__current_method.locals.keys())
        else:
            return []

    @property
    def __globals(self) -> List[str]:
        """
        Gets a list with the variables name in the global scope

        :return: A list with the variables names
        """
        # TODO: Global scope not implemented yet
        return []

    def get_symbol(self, identifier: str) -> ISymbol:
        """
        Gets a symbol in the symbol table by its id

        :param identifier: id of the symbol
        :return: the symbol if exists. Symbol None otherwise
        """
        if self.__current_method is not None and identifier in self.__current_method.symbols:
            return self.__current_method.symbols[identifier]
        elif identifier in self.symbol_table:
            return self.symbol_table[identifier]

        # the symbol may be a built in. If not, returns None
        symbol = Builtin.get_symbol(identifier)
        return symbol if symbol is not None else Type.none

    def convert_begin_method(self, method: Method):
        """
        Converts the signature of the method

        :param method: method that is being converted
        """
        num_args: int = len(method.args)
        num_vars: int = len(method.locals)

        init_data = bytearray([num_vars, num_args])
        self.__insert1(OpcodeInfo.INITSLOT, init_data)
        method.init_bytecode = self.last_code
        self.__current_method = method

        # updates each call to this method with its address
        if self.__is_entry_point_converted:
            for address, method in [call for call in self.__function_calls if call[1] is method]:
                self.__update_call(address, method)

    def convert_end_method(self):
        """
        Converts the end of the method
        """
        if self.last_code.opcode is not Opcode.RET:
            self.insert_return()
        if self.__current_method.is_main_method:
            self.__move_entry_point_to_beginning()
        self.__current_method = None

        self._stack.clear()

    def insert_return(self):
        """
        Insert the return statement
        """
        self.__insert1(OpcodeInfo.RET)

    def __move_entry_point_to_beginning(self):
        """
        Move the vm codes of the smart contract entry point to the beginning of the bytecode
        """
        method = self.__current_method
        if method is not None and method.is_main_method:
            main_first_code = self.__vm_codes_map[method.bytecode_address]
            main_last_code = self.last_code
            first_code = self.__vm_codes[0]

            if first_code.start_address != main_first_code.start_address:
                main_first_code._last_code = None
                first_code._last_code = main_last_code

                for vmcode, function in self.__function_calls:
                    if function.bytecode_address is not None:
                        self.__update_call(vmcode, function)

            self.__is_entry_point_converted = True

    def convert_begin_while(self) -> int:
        """
        Converts the beginning of the while statement

        :return: the address of the while first opcode
        """
        # it will be updated when the while ends
        self.__insert_jump(OpcodeInfo.JMP, 0)
        return self.last_code.start_address

    def convert_end_while(self, start_address: int, test_address: int):
        """
        Converts the end of the while statement

        :param start_address: the address of the while first opcode
        :param test_address: the address of the while test fist opcode
        """
        # updates the begin jmp with the target address
        self.__update_jump(start_address, test_address)

        # inserts end jmp
        while_begin: VMCode = self.__vm_codes_map[start_address]
        while_body: int = while_begin.end_address + 1
        end_jmp_to: int = while_body - self.address
        self.__insert_jump(OpcodeInfo.JMPIF, end_jmp_to)

    def convert_begin_if(self) -> int:
        """
        Converts the beginning of the if statement

        :return: the address of the if first opcode
        """
        # it will be updated when the if ends
        self.__insert_jump(OpcodeInfo.JMPIFNOT, 0)
        return self.last_code.start_address

    def convert_begin_else(self, start_address: int) -> int:
        """
        Converts the beginning of the if else statement

        :param start_address: the address of the if first opcode
        :return: the address of the if else first opcode
        """
        # it will be updated when the if ends
        self.__insert_jump(OpcodeInfo.JMP, 0)

        # updates the begin jmp with the target address
        self.__update_jump(start_address, self.address)

        return self.last_code.start_address

    def convert_end_if(self, start_address: int):
        """
        Converts the end of the if statement

        :param start_address: the address of the if first opcode
        """
        # updates the begin jmp with the target address
        self.__update_jump(start_address, self.address)

    def fix_negative_index(self):
        self.duplicate_stack_top_item()
        self.__insert1(OpcodeInfo.SIGN)
        self.convert_literal(-1)

        jmp_address = self.address
        self.__insert_jump(OpcodeInfo.JMPNE, 0)     # if index < 0

        self.duplicate_stack_item(2)                    # index += len(array)
        self.convert_builtin_method_call(Builtin.Len)
        self.convert_operation(BinaryOp.Add)

        self.__update_jump(jmp_address, self.address)

    def convert_literal(self, value: Any):
        """
        Converts a literal value

        :param value: the value to be converted
        """
        if isinstance(value, bool):
            self.convert_bool_literal(value)
        elif isinstance(value, int):
            self.convert_integer_literal(value)
        elif isinstance(value, str):
            self.convert_string_literal(value)
        elif value is None:
            self.insert_none()
        else:
            # TODO: convert other python literals as they are implemented
            raise NotImplementedError

    def convert_integer_literal(self, value: int):
        """
        Converts an integer literal value

        :param value: the value to be converted
        """
        if -1 <= value <= 16:
            opcode = Opcode.get_literal_push(value)
            if opcode is not None:
                op_info: OpcodeInformation = OpcodeInfo.get_info(opcode)
                self.__insert1(op_info)
        else:
            array = Integer(value).to_byte_array()
            self.convert_byte_array(array)
            # cast the value to integer
            self.__insert1(OpcodeInfo.CONVERT, StackItemType.Integer)
        self._stack.append(Type.int)

    def convert_string_literal(self, value: str):
        """
        Converts an string literal value

        :param value: the value to be converted
        """
        array = bytes(value, sys.getdefaultencoding())
        self.convert_byte_array(array)
        self._stack.pop()
        self._stack.append(Type.str)

    def convert_bool_literal(self, value: bool):
        """
        Converts an boolean literal value

        :param value: the value to be converted
        """
        if value:
            self.__insert1(OpcodeInfo.PUSH1)
        else:
            self.__insert1(OpcodeInfo.PUSH0)
        self._stack.append(Type.bool)

    def convert_byte_array(self, array: bytes):
        """
        Converts a byte value

        :param array: the value to be converted
        """
        data_len: int = len(array)
        if data_len <= ONE_BYTE_MAX_VALUE:
            op_info = OpcodeInfo.PUSHDATA1
        elif data_len <= TWO_BYTES_MAX_VALUE:
            op_info = OpcodeInfo.PUSHDATA2
        else:
            op_info = OpcodeInfo.PUSHDATA4

        data = Integer(data_len).to_byte_array(min_length=op_info.data_len) + array
        self.__insert1(op_info, data)
        self._stack.append(Type.none)  # TODO: change to bytearray when implemented

    def insert_none(self):
        """
        Converts None literal
        """
        self.__insert1(OpcodeInfo.PUSHNULL)
        self._stack.append(Type.none)

    def convert_new_empty_array(self, length: int, array_type: IType):
        """
        Converts the creation of a new empty array

        :param length: the size of the new array
        :param array_type: the Neo Boa type of the array
        """
        if length <= 0:
            self.__insert1(OpcodeInfo.NEWARRAY0)
        else:
            self.convert_literal(length)
            self.__insert1(OpcodeInfo.NEWARRAY)
        self._stack.append(array_type)

    def convert_new_array(self, length: int, array_type: IType):
        """
        Converts the creation of a new array

        :param length: the size of the new array
        :param array_type: the Neo Boa type of the array
        """
        if length <= 0:
            self.convert_new_empty_array(length, array_type)
        else:
            self.__insert1(OpcodeInfo.PACK)
            self._stack.append(array_type)

    def convert_set_new_array_item_at(self, index: int):
        """
        Converts the beginning of setting af a value in an array

        :param index: the index of the array that will be set
        """
        self.duplicate_stack_top_item()
        self.convert_literal_index(index)
        self.convert_literal(index)

    def convert_set_array_item(self):
        """
        Converts the end of setting af a value in an array
        """
        value_code = self.last_code
        start_address = self.address
        index_type: IType = self._stack[-2]  # top: index
        if index_type is Type.int:
            self.fix_negative_index()

        first_new_code = self.__vm_codes_map[start_address]
        first_new_code._last_code, value_code._last_code = value_code._last_code, self.last_code

        self.__insert1(OpcodeInfo.SETITEM)
        self._stack.pop()  # value
        self._stack.pop()  # index
        self._stack.pop()  # array

    def convert_get_array_item(self):
        """
        Converts the end of get a value in an array
        """
        index_type: IType = self._stack[-1]  # top: index
        if index_type is Type.int:
            self.fix_negative_index()

        array_type: IType = self._stack[-2]  # second-to-top: array
        if array_type is Type.str:
            self.convert_literal(1)  # length of substring
            self.convert_get_substring()
        else:
            self.__insert1(OpcodeInfo.PICKITEM)
            self._stack.pop()

    def convert_get_substring(self):
        """
        Converts the end of get a substring
        """
        self.__insert1(OpcodeInfo.SUBSTR)
        self._stack.pop()  # length
        self._stack.pop()  # start
        self._stack.pop()  # original string

    def convert_get_sub_array(self):
        """
        Converts the end of get a slice in the beginning of an array
        """
        # top: length, index, array
        if len(self._stack) > 2 and self._stack[-3] is Type.str:
            self.convert_get_substring()

    def convert_get_array_beginning(self):
        """
        Converts the end of get a slice in the beginning of an array
        """
        self.__insert1(OpcodeInfo.LEFT)
        self._stack.pop()  # length
        self._stack.pop()  # original array

    def convert_get_array_ending(self):
        """
        Converts the end of get a slice in the ending of an array
        """
        self.__insert1(OpcodeInfo.RIGHT)
        self._stack.pop()  # length
        self._stack.pop()  # original array

    def convert_load_symbol(self, symbol_id: str):
        """
        Converts the load of a symbol

        :param symbol_id: the symbol identifier
        """
        symbol = self.get_symbol(symbol_id)
        if symbol is not Type.none:
            if isinstance(symbol, Variable):
                self.convert_load_variable(symbol_id, symbol)
            elif isinstance(symbol, IBuiltinMethod) and symbol.opcode is not None:
                self.convert_builtin_method_call(symbol)
            else:
                self.convert_method_call(symbol)

    def convert_load_variable(self, var_id: str, var: Variable):
        """
        Converts the assignment of a variable

        :param var_id: the value to be converted
        :param var: the actual variable to be loaded
        """
        is_arg = False
        local = var_id in self.__current_method.symbols
        if local:
            is_arg = var_id in self.__args
            if is_arg:
                scope = self.__args
            else:
                scope = self.__locals
        else:
            scope = self.__globals

        index: int = scope.index(var_id)
        opcode = Opcode.get_load(index, local, is_arg)
        op_info = OpcodeInfo.get_info(opcode)

        if op_info.data_len > 0:
            self.__insert1(op_info, Integer(index).to_byte_array())
        else:
            self.__insert1(op_info)
        self._stack.append(var.type)

    def convert_store_variable(self, var_id: str):
        """
        Converts the assignment of a variable

        :param var_id: the value to be converted
        """
        is_arg = False
        local = var_id in self.__current_method.symbols
        if local:
            is_arg = var_id in self.__args
            if is_arg:
                scope = self.__args
            else:
                scope = self.__locals
        else:
            scope = self.__globals

        index: int = scope.index(var_id)
        opcode = Opcode.get_store(index, local, is_arg)
        op_info = OpcodeInfo.get_info(opcode)

        if op_info.data_len > 0:
            self.__insert1(op_info, Integer(index).to_byte_array())
        else:
            self.__insert1(op_info)
        self._stack.pop()

    def convert_builtin_method_call(self, function: IBuiltinMethod):
        """
        Converts a builtin method function call

        :param function: the function to be converted
        """
        if function.opcode is not None:
            op_info = OpcodeInfo.get_info(function.opcode)
            self.__insert1(op_info)
            for arg in function.args:
                self._stack.pop()
            self._stack.append(function.return_type)

    def convert_method_call(self, function: Method):
        """
        Converts a builtin method function call

        :param function: the function to be converted
        """
        if not self.__is_entry_point_converted or function.init_bytecode is None:
            self.__insert1(OpcodeInfo.CALL)
            self.__function_calls.append((self.last_code, function))
        else:
            function_address = Integer(function.bytecode_address - self.address)
            op_info = OpcodeInfo.CALL
            if len(function_address.to_byte_array(signed=True)) > op_info.max_data_len:
                op_info = OpcodeInfo.CALL_L

            data: bytes = function_address.to_byte_array(min_length=op_info.data_len, signed=True)
            self.__insert1(op_info, data)
        for arg in function.args:
            self._stack.pop()
        self._stack.append(function.return_type)

    def __update_call(self, vmcode: VMCode, method: Method):
        """
        Updates the data of a jump code in the bytecode

        :param vmcode: call code
        :param method: new data of the code
        """
        relative_address = Integer(method.bytecode_address - vmcode.start_address)
        op_info = vmcode.info
        if len(relative_address.to_byte_array(signed=True)) > op_info.max_data_len:
            op_info = OpcodeInfo.CALL_L

        vmcode.update(op_info, relative_address.to_byte_array(min_length=op_info.data_len, signed=True))
        self.__function_calls.remove((vmcode, method))

    def convert_operation(self, operation: IOperation):
        """
        Converts an operation

        :param operation: the operation that will be converted
        """
        for opcode, data in operation.opcode:
            op_info: OpcodeInformation = OpcodeInfo.get_info(opcode)
            self.__insert1(op_info, data)

        for op in range(operation.op_on_stack):
            self._stack.pop()
        self._stack.append(operation.result)

    def __insert1(self, op_info: OpcodeInformation, data: bytes = bytes()):
        """
        Inserts one opcode into the bytecode

        :param op_info: info of the opcode  that will be inserted
        :param data: data of the opcode, if needed
        """
        last_code = self.last_code
        vm_code = VMCode(op_info, last_code, data)

        self.__vm_codes_without_order.append(vm_code)

    def __insert_jump(self, op_info: OpcodeInformation, jump_to: int):
        """
        Inserts a jump opcode into the bytecode

        :param op_info: info of the opcode  that will be inserted
        :param jump_to: data of the opcode
        """
        if self.last_code.opcode is not Opcode.RET:
            op_info, data = self.__get_jump_data(op_info, jump_to)    # type:OpcodeInformation, bytes
            self.__insert1(op_info, data)
        for x in range(op_info.stack_items):
            self._stack.pop()

    def __update_jump(self, jump_address: int, updated_jump_to: int):
        """
        Updates the data of a jump code in the bytecode

        :param jump_address: jump code start address
        :param updated_jump_to: new data of the code
        """
        updated_jump_to -= jump_address
        vmcode: VMCode = self.__vm_codes_map[jump_address]
        if vmcode is not None:
            op_info, data = self.__get_jump_data(vmcode.info, updated_jump_to)  # type:OpcodeInformation, bytes
            vmcode.update(op_info, data)

    def __get_jump_data(self, op_info: OpcodeInformation, jump_to: int) -> Tuple[OpcodeInformation, bytes]:
        data: bytes = bytes()
        jmp_data: Integer = Integer(jump_to)

        if len(jmp_data.to_byte_array()) > op_info.max_data_len:
            opcode: Opcode = op_info.opcode.get_large_jump()
            if opcode is None:
                return op_info, data

            op_info = OpcodeInfo.get_info(opcode)
        data = jmp_data.to_byte_array(min_length=op_info.data_len, signed=True)

        return op_info, data

    def duplicate_stack_top_item(self):
        self.duplicate_stack_item()

    def duplicate_stack_item(self, pos: int = 1):
        """
        Duplicates the item n back in the stack

        :param pos: index of the variable
        """
        # n = 1 -> duplicates stack top item
        if pos > 0:
            opcode: Opcode = Opcode.get_dup(pos)
            if opcode is Opcode.PICK:
                self.convert_literal(pos)
            op_info = OpcodeInfo.get_info(opcode)
            self.__insert1(op_info)
            self._stack.append(self._stack[-pos])

    def swap_reverse_stack_items(self, no_items: int):
        if no_items > 1:
            opcode: Opcode = Opcode.get_reverse(no_items)
            if opcode is Opcode.REVERSEN:
                self.convert_literal(no_items)
            op_info = OpcodeInfo.get_info(opcode)
            self.__insert1(op_info)
            reverse = list(reversed(self._stack[-no_items:]))
            self._stack = self._stack[:-no_items]
            self._stack.extend(reverse)



from typing import Dict, Any

from boa3.model.type.anytype import anyType
from boa3.model.type.itype import IType
from boa3.model.type.primitive.booltype import BoolType
from boa3.model.type.primitive.inttype import IntType
from boa3.model.type.primitive.nonetype import NoneType
from boa3.model.type.primitive.strtype import StrType
from boa3.model.type.sequence.genericsequencetype import GenericSequenceType
from boa3.model.type.sequence.mutable.genericmutablesequencetype import GenericMutableSequenceType
from boa3.model.type.sequence.mutable.listtype import ListType
from boa3.model.type.sequence.tupletype import TupleType


class Type:
    @classmethod
    def values(cls) -> Dict[str, IType]:
        value_dict = {}
        for type in vars(cls).values():
            if isinstance(type, IType):
                value_dict[type._identifier] = type
        return value_dict

    @classmethod
    def get_type(cls, value: Any) -> IType:
        """
        Returns the type of the given value.

        :param value: value to get the type
        :return: Returns the type of the value. `Type.none` by default.
        """
        val: IType = None
        for type in vars(cls).values():
            if isinstance(type, IType) and type.is_type_of(value):
                val = type.build(value)
                break

        if val is not None:
            return val
        return cls.none

    @classmethod
    def get_generic_type(cls, type1: IType, type2: IType) -> IType:
        """
        Returns the common generic type between the given values.

        :param type1: first type to be compared
        :param type2: second type to be compared
        :return: Returns the common generic type of the values if exist. `Type.any` otherwise.
        """
        if type1 == type2 or type1.is_type_of(type2):
            return type1
        if type2.is_type_of(type1):
            return type2

        generic_types = [cls.mutableSequence, cls.sequence]
        for generic in generic_types:
            if generic.is_type_of(type1) and generic.is_type_of(type2):
                return generic
        return Type.any

    # Primitive Types
    int = IntType()
    bool = BoolType()
    str = StrType()
    none = NoneType()
    tuple = TupleType()
    list = ListType()

    # Generic types
    sequence = GenericSequenceType()
    mutableSequence = GenericMutableSequenceType()
    any = anyType



from abc import abstractmethod
from typing import Any

from boa3.model.symbol import ISymbol
from boa3.neo.vm.type.AbiType import AbiType


class IType(ISymbol):
    """
    An interface used to represent types

    :ivar identifier: the name identifier of the type
    """

    def __init__(self, identifier: str):
        self._identifier: str = identifier

    @property
    def shadowing_name(self) -> str:
        return 'type'

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def default_value(self) -> Any:
        return None

    @property
    def abi_type(self) -> AbiType:
        """
        Get the type representation for the abi

        :return: the representation for the abi. Any by default.
        """
        return AbiType.Any

    @classmethod
    @abstractmethod
    def _is_type_of(cls, value: Any) -> bool:
        """
        Validates if this is the type of the given value

        :param value: value to check the type
        :return: a boolean value that represents if this is the type of the value.
        :rtype: bool
        """
        pass

    def is_type_of(self, value: Any) -> bool:
        """
        Validates if this is the type of the given value. If this is not a primitive type, validates if its attributes
        are also valid.

        :param value: value to check the type
        :return: a boolean value that represents if this is the type of the value.
        :rtype: bool
        """
        return self._is_type_of(value)

    @classmethod
    @abstractmethod
    def build(cls, value: Any):
        """
        Creates a type instance with the given value

        :param value: value to build the type
        :return: The built type if the value is valid. None otherwise
        :rtype: IType or None
        """
        pass



import ast
from typing import List, Sequence, Union

from boa3 import helpers
from boa3.analyser.astanalyser import IAstAnalyser


class ConstructAnalyser(IAstAnalyser, ast.NodeTransformer):
    """
    This class is responsible for pre processing Python constructs

    The methods with the name starting with 'visit_' are implementations of methods from the :class:`NodeVisitor` class.
    These methods are used to walk through the Python abstract syntax tree.
    """

    def __init__(self, ast_tree: ast.AST):
        super().__init__(ast_tree)
        self.visit(self._tree)

    @property
    def tree(self) -> ast.AST:
        """
        Gets the analysed abstract syntax tree

        :return: the analysed ast
        """
        return self._tree

    def visit_For(self, for_node: ast.For):
        """
        Includes additional operations for converting the for statement into Neo VM

        :param for_node: the python ast for node
        """
        # get auxiliary variables
        var_iter_id = helpers.get_auxiliary_name(for_node, 'iter')
        var_index_id = helpers.get_auxiliary_name(for_node, 'index')

        add_instructs: List[ast.AST] = self.parse_to_node('{0} = x; {1} = 0'.format(var_iter_id, var_index_id), for_node)
        add_instructs[0].value = for_node.iter  # iter auxiliary variable is assigned with the node iter value

        update_target: ast.Assign = self.parse_to_node('x = {0}[{1}]'.format(var_iter_id, var_index_id), for_node)
        update_target.targets[0] = for_node.target  # target variable is updated each loop iteration

        update_index: ast.Assign = self.parse_to_node('{0} = {0} + 1'.format(var_index_id), for_node)
        loop_test: ast.Compare = self.parse_to_node('{0} < len({1})'.format(var_index_id, var_iter_id), for_node)

        for_node.body.insert(0, update_target)
        for_node.body.append(update_index)
        for_node.body.append(loop_test)

        for node in add_instructs:
            self.generic_visit(node)
        self.generic_visit(for_node)

        add_instructs.append(for_node)
        return add_instructs

    def parse_to_node(self, expression: str, origin: ast.AST = None) -> Union[ast.AST, Sequence[ast.AST]]:
        """
        Parses an expression to an ast.

        :param expression: string expression to be parsed
        :param origin: an existing ast. If not None, the parsed node will have the same location of origin.
        :return: the parsed node
        :rtype: ast.AST or Sequence[ast.AST]
        """
        node = ast.parse(expression)
        if origin is not None:
            self.update_line_and_col(node, origin)

        # get the expression instead of the default root node
        if hasattr(node, 'body'):
            node = node.body
        elif hasattr(node, 'argtypes'):
            node = node.argtypes

        if isinstance(node, list) and len(node) == 1:
            # the parsed node has a list of expression and only one expression is found
            result = node[0]
        else:
            result = node

        if isinstance(result, ast.Expr):
            # an expr node encapsulates another node in its value field.
            result = result.value
        return result

    def update_line_and_col(self, target: ast.AST, origin: ast.AST):
        """
        Updates the position of a node and its child nodes

        :param target: the node that will have its position updated
        :param origin: the node with the desired position
        """
        ast.copy_location(target, origin)
        for field, value in ast.iter_fields(target):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast.AST):
                        ast.copy_location(item, origin)
            elif isinstance(value, ast.AST):
                ast.copy_location(value, origin)
        ast.fix_missing_locations(target)



from typing import Optional

from boa3.model.expression import IExpression
from boa3.model.type.itype import IType


class Variable(IExpression):
    """
    A class used to represent a variable

    :ivar var_type: the type of the variable.
    """

    def __init__(self, var_type: Optional[IType]):
        self.__var_type: Optional[IType] = var_type

    @property
    def shadowing_name(self) -> str:
        return 'variable'

    @property
    def type(self) -> IType:
        return self.__var_type

    def set_type(self, var_type: IType):
        """
        Sets a type for the variable if its type is not defined

        :param var_type:
        """
        if self.__var_type is None:
            self.__var_type = var_type



from abc import abstractmethod

from boa3.model.symbol import ISymbol
from boa3.model.type.itype import IType


class IExpression(ISymbol):
    """
    An interface used to represent expressions
    """

    @property
    @abstractmethod
    def type(self) -> IType:
        """
        Gets the type of the evaluated expression

        :return: the resulting type when the expression is evaluated
        """
        pass



from enum import Enum


class AbiType(str, Enum):
    Signature = 'Signature'
    Boolean = 'Boolean'
    Integer = 'Integer'
    Hash160 = 'Hash160'
    Hash256 = 'Hash256'
    ByteArray = 'ByteArray'
    PublicKey = 'PublicKey'
    String = 'String'
    Array = 'Array'
    Map = 'Map'
    InteropInterface = 'InteropInterface'
    Any = 'Any'
    Void = 'Void'



from typing import Optional

from boa3.model.builtin.classmethod.appendmethod import AppendMethod
from boa3.model.builtin.decorator.builtindecorator import IBuiltinDecorator
from boa3.model.builtin.decorator.publicdecorator import PublicDecorator
from boa3.model.builtin.method.lenmethod import LenMethod
from boa3.model.method import Method
from boa3.model.type.itype import IType


class Builtin:
    @classmethod
    def get_symbol(cls, symbol_id: str) -> Optional[Method]:
        for name, method in vars(cls).items():
            if isinstance(method, IBuiltinDecorator) and method.identifier == symbol_id:
                return method

    @classmethod
    def get_by_self(cls, symbol_id: str, self_type: IType) -> Optional[Method]:
        from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
        for name, method in vars(cls).items():
            if (isinstance(method, IBuiltinMethod)
                    and method.identifier == symbol_id
                    and method.validate_self(self_type)):
                return method

    # builtin method
    Len = LenMethod()

    # python class method
    Append = AppendMethod()

    # builtin decorator
    Public = PublicDecorator()



from abc import ABC, abstractmethod
from typing import List, Any, Iterable

from boa3.model.type.itype import IType
from boa3.neo.vm.type.AbiType import AbiType


class SequenceType(IType, ABC):
    """
    An interface used to represent Python sequence type
    """

    def __init__(self, identifier: str, values_type: List[IType]):
        self.value_type: IType = self.__initialize_sequence_type(values_type)
        super().__init__(identifier)

    @property
    def identifier(self) -> str:
        return '{0}[{1}]'.format(self._identifier, self.value_type.identifier)

    @property
    def default_value(self) -> Any:
        return []

    def is_type_of(self, value: Any) -> bool:
        if self._is_type_of(value):
            if isinstance(value, SequenceType):
                return self.value_type.is_type_of(value.value_type)
            return True
        return False

    @classmethod
    def get_types(cls, value: Any) -> List[IType]:
        from boa3.model.type.type import Type
        if isinstance(value, IType):
            return [value]

        if not isinstance(value, Iterable):
            value = [value]

        types: List[IType] = [val if isinstance(val, IType) else Type.get_type(val) for val in value]
        types = list(set(types))  # get list of different types in the tuple
        return cls.filter_types(types)

    @property
    def abi_type(self) -> AbiType:
        return AbiType.Array  # TODO: change when 'bytes' is implemented

    @abstractmethod
    def is_valid_key(self, value_type: IType) -> bool:
        """
        Verifies if the given type is valid to retrieve values

        :param value_type: type to validate.
        :return: A boolean value that represents if the value is valid.
        """
        return False

    @property
    def can_reassign_values(self) -> bool:
        return True

    @property
    @abstractmethod
    def valid_key(self) -> IType:
        pass

    def __initialize_sequence_type(self, values_type: List[IType]):
        if len(values_type) != 1:
            from boa3.model.type.anytype import anyType
            val_type: IType = anyType
        else:
            val_type: IType = values_type[0]

        return val_type

    @classmethod
    def filter_types(cls, values_type) -> List[IType]:
        if values_type is None:
            values_type = []
        elif not isinstance(values_type, list):
            values_type = [values_type]

        if len(values_type) > 1:
            from boa3.model.type.type import Type
            if any(t is Type.any or t is Type.none for t in values_type):
                return [Type.any]

            # verifies if all the types are the same sequence with different arguments
            if all(isinstance(x, SequenceType) for x in values_type):
                sequence_type = type(values_type[0])  # first sequence type
                value_type = type(values_type[0].value_type)  # first sequence values type

                if all(isinstance(x, sequence_type) for x in values_type):
                    # if all the types are the same sequence type, build this sequence with any as parameters
                    types = list(set(value.value_type for value in values_type))
                    values_type = [sequence_type(values_type=types)]
                elif all(isinstance(x.value_type, value_type) for x in values_type):
                    # the sequences doesn't have the same type but the value type is the same
                    # for example: Tuple[int] and List[int]
                    values_type = [Type.sequence.build_sequence(values_type[0].value_type)]
                else:
                    # otherwise, built a generic sequence with any as parameters
                    values_type = [Type.sequence]
        return values_type

    @classmethod
    def build_sequence(cls, value_type: IType):
        """
        Creates a sequence type instance with the given value

        :param value_type: value to build the type
        :return: The built sequence type.
        :rtype: IType or None
        """
        return cls(values_type=[value_type])  # get list of different types in the tuple



from boa3.compiler.compiler import Compiler
from boa3.exception.InvalidPathException import InvalidPathException


class Boa3:
    """
    The main class.
    Contains the methods that the final user have access to.
    """

    @staticmethod
    def compile(path: str) -> bytes:
        """
        Load a Python file to be compiled but don't write the result into a file

        :param path: the path of the Python file to compile
        :return: the bytecode of the compiled .nef file
        """
        if not path.endswith('.py'):
            raise InvalidPathException(path)

        return Compiler().compile(path)

    @staticmethod
    def compile_and_save(path: str, output_path: str = None):
        """
        Load a Python file to be compiled and save the result into the files.
        By default, the resultant .nef file is saved in the same folder of the
        source file.

        :param path: the path of the Python file to compile
        :param output_path: Optional path to save the generated files
        """
        if not path.endswith('.py'):
            raise InvalidPathException(path)

        if output_path is None:
            output_path = path.replace('.py', '.nef')
        elif not output_path.endswith('.nef'):
            raise InvalidPathException(path)

        Compiler().compile_and_save(path, output_path)



import sys

from boa3.constants import SIZE_OF_INT32
from boa3.neo.vm.type.Integer import Integer


class Serializer:
    """
    This class is responsible for serialize Python primitive types into bytes
    """

    def __init__(self):
        self.__result = bytearray()

    @property
    def result(self) -> bytearray:
        """
        Gets the serializer written data

        :return: the written data as a bytearray
        :rtype: bytearray
        """
        return self.__result

    def write_bytes(self, value: bytes, size: int = 0):
        """
        Write the given bytes object into the serializer output

        :param value: bytes to be written
        :param size: the number of bytes that must filled
        """
        # if the number of bytes is greater than the specified size, the remaining will be ignored
        if 0 < size < len(value):
            value = value[0:size]

        self.__result.extend(value)
        # if the number of bytes is greater than the specified, fill the remaining with an empty bytearray
        if size > len(value):
            self.__fill_with_empty(size - len(value))

    def write_byte_array(self, value: bytearray, size: int = 0):
        """
        Write the given bytearray object into the serializer output

        :param value: bytes to be written
        :param size: the number of bytes that must filled
        """
        # if the number of bytes is greater than the specified size, the remaining will be ignored
        if 0 < size < len(value):
            value = value[0:size]

        self.__result.extend(value)
        # if the number of bytes is greater than the specified, fill the remaining with an empty bytearray
        if size > len(value):
            self.__fill_with_empty(size - len(value))

    def write_string(self, string: str, size: int):
        """
        Write the given string into the serializer output

        :param string: string to be written
        :param size: the number of bytes that the string must fill
        """
        # if the size of the string is greater than the specified size, the remaining will be ignored
        if len(string) > size:
            string = string[0:size]

        str_bytes = bytearray(string, encoding='UTF-8')
        self.__result.extend(str_bytes)

        # if the size of the string is less than the specified, fill the remaining with an empty bytearray
        if len(string) < size:
            self.__fill_with_empty(size - len(string))

    def write_integer(self, integer: int, size: int = SIZE_OF_INT32):
        """
        Write the given integer value into the serializer output

        :param integer: integer value to be written
        :param size: the number of bytes that the string must fill
        """
        int_bytes = integer.to_bytes(size, sys.byteorder)
        self.__result.extend(int_bytes)

    def write_value(self, var_bytes: bytes):
        """
        Write the given bytes object and size into the serializer output

        :param var_bytes: bytes to be written
        """
        value_bytes = bytearray()
        value_bytes.extend(Integer(len(var_bytes)).to_byte_array())
        value_bytes.extend(var_bytes)
        self.__result.extend(value_bytes)

    def __fill_with_empty(self, length: int):
        """
        Write an empty array with the given length into the serializer output

        :param length: the number of bytes that must be filled
        """
        self.__result.extend(bytes(length))



import logging
import os

from boa3.analyser.analyser import Analyser
from boa3.compiler.codegenerator import CodeGenerator
from boa3.compiler.filegenerator import FileGenerator
from boa3.exception.NotLoadedException import NotLoadedException


class Compiler:
    """
    The main compiler class.

    :ivar bytecode: the compiled file as a byte array. Empty by default.
    """

    def __init__(self):
        self.bytecode: bytearray = bytearray()
        self.__analyser: Analyser = None

    def compile(self, path: str) -> bytes:
        """
        Load a Python file and tries to compile it

        :param path: the path of the Python file to compile
        :return: the bytecode of the compiled .nef file
        """
        fullpath = os.path.realpath(path)
        filepath, filename = os.path.split(fullpath)

        logging.info('Started compiling\t{0}'.format(filename))
        self.__analyse(path)
        return self.__compile()

    def compile_and_save(self, path: str, output_path: str):
        """
        Save the compiled file and the metadata files

        :param path: the path of the Python file to compile
        :param output_path: the path to save the generated files
        """
        self.bytecode = self.compile(path)
        self.__save(output_path)

    def __analyse(self, path: str):
        """
        Load a Python file and analyses its syntax

        :param path: the path of the Python file to compile
        """
        self.__analyser = Analyser.analyse(path)

    def __compile(self) -> bytes:
        """
        Compile the analysed Python file.

        :return: the compiled file as a bytecode.
        :raise NotLoadedException: raised if none file were analysed
        """
        if not self.__analyser.is_analysed:
            raise NotLoadedException
        return CodeGenerator.generate_code(self.__analyser)

    def __save(self, output_path: str):
        """
        Save the compiled file and the metadata files

        :param output_path: the path to save the generated files
        :raise NotLoadedException: raised if no file were compiled
        """
        if (self.__analyser is None
                or not self.__analyser.is_analysed
                or len(self.bytecode) == 0):
            raise NotLoadedException

        generator = FileGenerator(self.bytecode, self.__analyser.symbol_table)

        nef_bytes = generator.generate_nef_file()
        with open(output_path, 'wb+') as nef_file:
            nef_file.write(nef_bytes)
            nef_file.close()

        manifest_path = output_path.replace('.nef', '.manifest.json')
        manifest_bytes = generator.generate_manifest_file()
        with open(manifest_path, 'wb+') as manifest_file:
            manifest_file.write(manifest_bytes)
            manifest_file.close()



import re
import sys

import boa3
from boa3 import neo
from boa3.constants import SIZE_OF_INT32, SIZE_OF_INT160, DEFAULT_UINT32
from boa3.neo import cryptography
from boa3.neo.utils.serializer import Serializer
from boa3.neo.vm.type.Integer import Integer


class NefFile:
    """
    The object encapsulates the information of the NEO Executable Format (NEF)

    :ivar compiler: the name of the compiler used to generate the file
    :ivar version: the version of the compiler
    :ivar script_hash: the smart contract hash
    :ivar check_sum: the check sum of the file.
    :ivar script: the script of the smart contract
    """
    # Constants
    __COMPILER_HEADER_SIZE = 32
    __VERSION_NUMBER_OF_FIELDS = 4  # major.minor.patch-release
    __VERSION_HEADER_SIZE = __VERSION_NUMBER_OF_FIELDS * SIZE_OF_INT32

    def __init__(self, script_bytes: bytes):
        """
        :param script_bytes: the script of the smart contract
        """
        self.__magic: int = 0x3346454E              # NEO Executable Format 3 (NEF3)
        self.compiler: str = "neo3-boa by COZ"      # Compiler Name
        self.version: str = boa3.__version__        # Compiler Version
        self.check_sum: int = 0

        self.script: bytes = script_bytes           # Smart Contract Script
        self.script_hash: bytes = neo.to_script_hash(self.script)  # Script Hash
        self.check_sum = self.compute_check_sum()   # Checksum

    @property
    def size(self) -> int:
        """
        Get the size in bytes of the NEF file

        :return: size of NEF File.
        """
        # it is calculated with constants to be compatible with Neo code
        # because Python sizeof differs from C# sizeof
        return (
            SIZE_OF_INT32                       # size of magic
            + NefFile.__COMPILER_HEADER_SIZE    # total size of compiler's name
            + NefFile.__VERSION_HEADER_SIZE     # total size of compiler's version
            + SIZE_OF_INT160                    # size of script hash
            + SIZE_OF_INT32                     # size of check sum
            + len(self.script_len_in_bytes)     # size used to store script size
            + len(self.script)                  # size of smart contract script
        )

    @property
    def script_len_in_bytes(self) -> bytes:
        """
        Get the size of the script in bytes to compute the size of the header

        :return: size of the script in bytes.
        """
        return Integer(len(self.script)).to_byte_array()

    @property
    def __header_size_before_check_sum(self) -> int:
        """
        Get the size of the header of the NEF file before the checksum

        :return: size of header.
        """
        return (
            SIZE_OF_INT32                       # size of magic
            + NefFile.__COMPILER_HEADER_SIZE    # total size of compiler's name
            + NefFile.__VERSION_HEADER_SIZE     # total size of compiler's version
            + SIZE_OF_INT160                    # size of script hash
        )

    @property
    def __version_info(self) -> list:
        """
        Returns the information about the compiler version

        :return: a list of the fields of the version
        """
        version_info = re.split('[.-]', self.version)  # major.minor.patch-release

        for index, field in enumerate(version_info):
            try:
                version_info[index] = int(field)
            except ValueError:
                # in python versions, the release field is a string
                # the nef header needs int values in the version
                version_info[index] = DEFAULT_UINT32

        while len(version_info) < NefFile.__VERSION_NUMBER_OF_FIELDS:
            version_info.append(DEFAULT_UINT32)

        return version_info[0:NefFile.__VERSION_NUMBER_OF_FIELDS]

    def serialize(self) -> bytes:
        """
        Serialize the NefFile object

        :return: bytes of the serialized object.
        """
        nef_serializer = Serializer()

        nef_serializer.write_integer(self.__magic)
        nef_serializer.write_string(self.compiler, NefFile.__COMPILER_HEADER_SIZE)

        for info in self.__version_info:
            nef_serializer.write_integer(info)

        nef_serializer.write_bytes(self.script_hash)
        nef_serializer.write_integer(self.check_sum)
        nef_serializer.write_value(self.script)

        return nef_serializer.result

    def compute_check_sum(self) -> int:
        """
        Computes the checksum of the NEF file

        :return: computed check sum.
        """
        serialized = self.serialize()
        size = self.__header_size_before_check_sum
        check_sum = cryptography.sha256(serialized[0:size])
        return int.from_bytes(check_sum[0:SIZE_OF_INT32], sys.byteorder, signed=False)



from boa3.neo.vm.opcode.Opcode import Opcode


class OpcodeInformation:
    """
    Stores the information about a specific :class:`Opcode`

    :ivar opcode: the opcode of the code
    :ivar data_len: the size in bytes of the expected operand. Zero by default.
    :ivar max_data_len: the max size in bytes of the operand. Same value as data_len if size is constant.
    """

    def __init__(self, opcode: Opcode, min_data_len: int = 0, extra_data_max_len: int = 0, stack_items: int = 0):
        self.opcode: Opcode = opcode

        if min_data_len < 0:
            min_data_len = 0
        self.data_len: int = min_data_len

        if extra_data_max_len < 0:
            extra_data_max_len = 0
        self.max_data_len: int = min_data_len + extra_data_max_len

        if stack_items < 0:
            stack_items = 0
        self.stack_items: int = stack_items

    def get_large(self):
        large_op = self.opcode.get_large
        if large_op is None:
            return None

        from boa3.neo.vm.opcode.OpcodeInfo import OpcodeInfo
        return OpcodeInfo.get_info(large_op)



from typing import Dict, Optional

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.sequence.sequencetype import SequenceType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class LenMethod(IBuiltinMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        identifier = 'len'
        args: Dict[str, Variable] = {'__o': Variable(Type.sequence)}
        super().__init__(identifier, args, Type.int)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        if not isinstance(params[0], IExpression):
            return False
        return isinstance(params[0].type, SequenceType)

    @property
    def opcode(self) -> Optional[Opcode]:
        return Opcode.SIZE

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None



from abc import ABC, abstractmethod
from typing import Any, Dict, Optional

from boa3.model.builtin.decorator.builtindecorator import IBuiltinDecorator
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class IBuiltinMethod(IBuiltinDecorator, ABC):
    def __init__(self, identifier: str, args: Dict[str, Variable] = None, return_type: IType = None):
        self.identifier = identifier
        super().__init__(identifier, args, return_type)

    @property
    def opcode(self) -> Optional[Opcode]:
        """
        Gets the opcode for the method.

        :return: the opcode if exists. None otherwise.
        """
        return None

    @property
    def args_on_stack(self) -> int:
        """
        Gets the number of arguments that must be on stack before the opcode is called.

        :return: the number of arguments if opcode is not None. Zero otherwise.
        """
        if self.opcode is None:
            return 0
        else:
            num_args = self._args_on_stack
            if num_args < 0:
                return 0
            elif num_args > len(self.args):
                return len(self.args)
            return num_args

    def push_self_first(self) -> bool:
        """
        Verifies if the `self` value of the method needs to be pushed to the Neo execution stack before the
        other arguments.

        :return: a boolean value indicating if the `self` argument must be pushed before. Returns False if there isn't
                 a `self` argument in the function
        """
        return False

    def validate_self(self, self_type: IType) -> bool:
        """
        Verifies if the given value is valid to the function `self` argument

        :param self_type: type of the value
        :return: a boolean value that represents if the value is valid. Returns False if there isn't a `self` argument
                 in the function
        """
        if not self.has_self_argument:
            return False
        return self.args['self'].type.is_type_of(self_type)

    @property
    def has_self_argument(self) -> bool:
        """
        Verifies if the function has a `self` argument.

        :return: True if there is this argument. False otherwise.
        """
        return 'self' in self.args

    @property
    @abstractmethod
    def _args_on_stack(self) -> int:
        """
        Gets the number of arguments that must be on stack before the opcode is called.

        :return: the number of arguments.
        """
        return 0

    @property
    def body(self) -> Optional[str]:
        """
        Gets the body of the method.

        :return: Return the code of the method body if there is no opcode. None otherwise.
        """
        if self.opcode is None:
            return self._body
        else:
            return None

    @property
    @abstractmethod
    def _body(self) -> Optional[str]:
        """
        Gets the body of the method.

        :return: Return the code of the method body.
        """
        return None

    def build(self, value: Any):
        """
        Creates a method instance with the given value as self

        :param value: value to build the type
        :return: The built method if the value is valid. The current object otherwise
        :rtype: IBuiltinMethod
        """
        return self



import ast
import logging
from typing import Any, Dict, List, Optional, Tuple, Union

from boa3 import helpers
from boa3.analyser.astanalyser import IAstAnalyser
from boa3.exception import CompilerError, CompilerWarning
from boa3.model.builtin.builtin import Builtin
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.method import Method
from boa3.model.module import Module
from boa3.model.symbol import ISymbol
from boa3.model.type.itype import IType
from boa3.model.type.sequence.sequencetype import SequenceType
from boa3.model.type.type import Type
from boa3.model.variable import Variable


class ModuleAnalyser(IAstAnalyser, ast.NodeVisitor):
    """
    This class is responsible for mapping the locals of the functions and modules

    The methods with the name starting with 'visit_' are implementations of methods from the :class:`NodeVisitor` class.
    These methods are used to walk through the Python abstract syntax tree.

    :ivar modules: a dictionary that maps each module with its name. Empty by default.
    :ivar symbols: a dictionary that maps the global symbols.
    """

    def __init__(self, ast_tree: ast.AST, symbol_table: Dict[str, ISymbol]):
        super().__init__(ast_tree)
        self.modules: Dict[str, Module] = {}
        self.symbols: Dict[str, ISymbol] = symbol_table

        self.__builtin_functions_to_visit: Dict[str, IBuiltinMethod] = {}
        self.__current_module: Module = None
        self.__current_method: Method = None

        self.visit(self._tree)

    @property
    def __current_scope(self):
        """
        Returns the scope that is currently being analysed

        :return: the current scope. Return None if it is the global scope
        :rtype: Method or Module or None
        """
        if self.__current_method is not None:
            return self.__current_method
        return self.__current_module

    @property
    def global_symbols(self) -> Dict[str, ISymbol]:
        """
        Returns the symbols in global scope

        :return: the symbol table of the global scope
        """
        global_symbols: Dict[str, ISymbol] = {}

        global_symbols.update(self.symbols)
        for mod in self.modules.values():
            global_symbols.update(mod.symbols)

        return global_symbols

    def __include_variable(self, var_id: str, var_type_id: Union[str, IType], source_node: ast.AST, var_enumerate_type: IType = Type.none):
        """
        Includes the variable in the symbol table if the id was not used

        :param var_id: variable id
        :param var_type_id: variable type id
        :type var_type_id: str or IType
        :param var_enumerate_type: variable value type id if var_type_id is a SequenceType
        """
        if var_id not in self.__current_scope.symbols:
            outer_symbol = self.get_symbol(var_id)
            if outer_symbol is not None:
                self._log_warning(
                    CompilerWarning.NameShadowing(source_node.lineno, source_node.col_offset, outer_symbol, var_id)
                )

            var_type = None
            if isinstance(var_type_id, SequenceType):
                var_type = var_type_id
                var_enumerate_type = self.get_enumerate_type(var_type)
            else:
                if isinstance(var_type_id, IType):
                    var_type_id = var_type_id.identifier
                    var_type = self.symbols[var_type_id]

            # when setting a None value to a variable, set the variable as any type
            if var_type is Type.none:
                var_type = Type.any

            if isinstance(var_type, IType) or var_type is None:
                # if type is None, the variable type depends on the type of a expression
                if isinstance(var_type, SequenceType):
                    var_type = var_type.build_sequence(var_enumerate_type)
                var = Variable(var_type)
                self.__current_scope.include_variable(var_id, var)

    def __include_method(self, method_id: str, method: Method):
        """
        Includes the method in the symbol table if the id was not used

        :param method_id: method id
        :param method: method to be included
        """
        if method_id not in self.__current_module.symbols:
            self.__current_module.include_method(method_id, method)

    def get_symbol(self, symbol_id: str) -> Optional[ISymbol]:
        if symbol_id in self.__current_scope.symbols:
            # the symbol exists in the local scope
            return self.__current_scope.symbols[symbol_id]
        elif symbol_id in self.__current_module.symbols:
            # the symbol exists in the module scope
            return self.__current_module.symbols[symbol_id]
        else:
            return super().get_symbol(symbol_id)

    def visit_Module(self, module: ast.Module):
        """
        Visitor of the module node

        Fills module symbol table

        :param module:
        """
        mod: Module = Module()
        self.__current_module = mod
        for stmt in module.body:
            self.visit(stmt)
        # TODO: include the body of the builtin methods to the ast
        # TODO: get module name
        self.modules['main'] = mod
        self.__current_module = None

    def visit_FunctionDef(self, function: ast.FunctionDef):
        """
        Visitor of the function node

        Includes the method in the scope of its module

        :param function:
        """
        fun_args = self.visit(function.args)
        fun_rtype_symbol = self.visit(function.returns) if function.returns is not None else Type.none

        if fun_rtype_symbol is None:
            # it is a function with None return: Main(a: int) -> None:
            raise NotImplementedError

        if isinstance(fun_rtype_symbol, str):
            symbol = self.get_symbol(function.returns.id.lower())
            fun_rtype_symbol = self.get_type(symbol)

        fun_decorators = [self.visit(decorator) for decorator in function.decorator_list]
        fun_return: IType = self.get_type(fun_rtype_symbol)
        method = Method(fun_args, fun_return, Builtin.Public.identifier in fun_decorators)

        if function.name in ['main', 'Main']:
            # don't show return when the function is void
            return_type = ' -> {0}'.format(fun_return.identifier) if fun_return is not Type.none else ''
            arg_types = ', '.join(['{0}: {1}'.format(arg, var.type.identifier) for arg, var in fun_args.items()])

            logging.info("Main method found at {0}:{1}: '{2}({3}){4}'"
                         .format(function.lineno, function.col_offset, function.name, arg_types, return_type))
            # main method is always public
            method.set_as_main_method()

        self.__current_method = method

        for stmt in function.body:
            self.visit(stmt)

        self.__include_method(function.name, method)
        self.__current_method = None

    def visit_arguments(self, arguments: ast.arguments) -> Dict[str, Variable]:
        """
        Visitor of the function arguments node

        :param arguments:
        :return: a dictionary that maps each argument to its identifier
        """
        args: Dict[str, Variable] = {}

        for arg in arguments.args:
            var_id, var = self.visit_arg(arg)   # Tuple[str, Variable]
            args[var_id] = var
        return args

    def visit_arg(self, arg: ast.arg) -> Tuple[str, Variable]:
        """
        Visitor of a function argument node

        :param arg:
        :return: a tuple with the identifier and the argument
        """
        var_id = arg.arg
        var_type: IType = self.get_type(arg.annotation)

        if var_type is Type.none and isinstance(arg.annotation, ast.Name):
            var_symbol: ISymbol = self.get_symbol(arg.annotation.id.lower())
            var_type = self.get_type(var_symbol)

        return var_id, Variable(var_type)

    def visit_Return(self, ret: ast.Return):
        """
        Visitor of the function return node

        If the return is a name, verifies if the symbol is defined

        :param ret: the python ast return node
        """
        if isinstance(ret.value, ast.Name):
            symbol_id = self.visit(ret.value)
            symbol = self.get_symbol(symbol_id)
            if symbol is None:
                # the symbol doesn't exists
                self._log_error(
                    CompilerError.UnresolvedReference(ret.value.lineno, ret.value.col_offset, symbol_id)
                )

    def visit_type(self, target: ast.AST) -> Optional[IType]:
        """
        Gets the type by its identifier

        :param target: ast node to be evaluated
        :return: the type of the value inside the node. None by default
        """
        target_type = self.visit(target)  # Type:str or IType
        if isinstance(target_type, str) and not isinstance(target, ast.Str):
            symbol = self.get_symbol(target_type)
            if symbol is None:
                symbol = self.get_symbol(target_type.lower())
            if symbol is None:
                # the symbol doesn't exists
                self._log_error(
                    CompilerError.UnresolvedReference(target.lineno, target.col_offset, target_type)
                )
            target_type = symbol

        if target_type is None and not isinstance(target, ast.NameConstant):
            # the value type is invalid
            return None
        return self.get_type(target_type)

    def get_enumerate_type(self, var_type: IType) -> IType:
        return var_type.value_type if isinstance(var_type, SequenceType) else Type.none

    def visit_Assign(self, assign: ast.Assign):
        """
        Visitor of the variable assignment node

        Includes the variable in its scope if it's the first use

        :param assign:
        """
        var_id = self.visit(assign.targets[0])
        var_type = self.visit_type(assign.value)

        self.__include_variable(var_id, var_type, source_node=assign)

    def visit_AnnAssign(self, ann_assign: ast.AnnAssign):
        """
        Visitor of the annotated variable assignment node

        Includes the variable in its scope if it's the first use

        :param ann_assign:
        """
        var_id: str = self.visit(ann_assign.target)
        var_type: IType = self.visit_type(ann_assign.annotation)

        # TODO: check if the annotated type and the value type are the same
        self.__include_variable(var_id, var_type, source_node=ann_assign)

    def visit_Expr(self, expr: ast.Expr):
        """
        Visitor of the atom expression node

        :param expr:
        """
        value = self.visit(expr.value)
        if isinstance(expr.value, ast.Name):
            if (
                # it is not a symbol of the method scope
                (self.__current_method is not None and value not in self.__current_method.symbols)
                # nor it is a symbol of the module scope
                or (self.__current_module is not None and value not in self.__current_module.symbols)
                # nor it is a symbol of the global scope
                or value not in self.symbols
            ):
                self._log_error(
                    CompilerError.UnresolvedReference(expr.value.lineno, expr.value.col_offset, value)
                )

    def visit_Subscript(self, subscript: ast.Subscript):
        """
        Verifies if it is the types in the subscription are valid

        :param subscript: the python ast subscription node
        :return: if the subscript is not a symbol, returns its type. Otherwise returns the symbol id.
        :rtype: IType or str
        """
        value = self.visit(subscript.value)
        symbol = self.get_symbol(value)

        if symbol is None:
            symbol = self.get_symbol(value.lower())
            subscript.value.id = value.lower() if symbol is not None else subscript.value

        if isinstance(subscript.ctx, ast.Load):
            if isinstance(symbol, SequenceType):
                values_type: IType = self.get_values_type(subscript.slice.value)
                return symbol.build_sequence(values_type)

            symbol_type = self.get_type(symbol)
            if isinstance(symbol_type, SequenceType):
                return symbol_type.value_type

        return value

    def get_values_type(self, value: ast.AST):
        """
        Verifies if it is a multiple assignments statement

        :param value: the python ast subscription node
        """
        value_type: Optional[IType] = None

        if isinstance(value, ast.Subscript):
            # index is another subscription
            value_type = self.visit(value)
        elif isinstance(value, ast.Name):
            # index is an identifier
            value_type = self.get_symbol(value.id)
            if value_type is None:
                value_type = self.get_symbol(value.id.lower())
                value.id = value.id.lower() if value_type is not None else value.id

        if not isinstance(value_type, IType):
            # type hint not using identifiers or using identifiers that are not types
            index = self.visit(value)
            self._log_error(
                CompilerError.UnresolvedReference(value.lineno, value.col_offset, index)
            )

        return value_type

    def visit_Call(self, call: ast.Call) -> Optional[IType]:
        """
        Visitor of a function call node

        :param call: the python ast function call node
        :return: the result type of the called function. None if the function is not found
        """
        func_id = self.visit(call.func)
        func_symbol = self.get_symbol(func_id)

        # if func_symbol is None, the called function may be a function written after in the code
        # that's why it shouldn't log a compiler error here
        if func_symbol is None:
            return None

        if not isinstance(func_symbol, Method):
            # the symbol doesn't exists
            self._log_error(
                CompilerError.UnresolvedReference(call.func.lineno, call.func.col_offset, func_id)
            )
        elif isinstance(func_symbol, IBuiltinMethod) and func_symbol.body is not None:
            self.__builtin_functions_to_visit[func_id] = func_symbol

        return self.get_type(call.func)

    def visit_For(self, for_node: ast.For):
        """
        Visitor of for statement node

        :param for_node: the python ast for node
        """
        iter_type = self.get_type(for_node.iter)
        targets = self.visit(for_node.target)

        if isinstance(iter_type, SequenceType):
            if isinstance(targets, str):
                self.__include_variable(targets, iter_type.value_type, source_node=for_node.target)
            else:
                for target in targets:
                    if isinstance(target, str):
                        self.__include_variable(target, iter_type.value_type, source_node=for_node.target)

            for_iter_id = helpers.get_auxiliary_name(for_node, 'iter')
            for_index_id = helpers.get_auxiliary_name(for_node, 'index')

            self.__include_variable(for_iter_id, iter_type, source_node=for_node.iter)
            self.__include_variable(for_index_id, iter_type.valid_key, source_node=for_node)

        # continue to walk through the tree
        self.generic_visit(for_node)

    def visit_Name(self, name: ast.Name) -> str:
        """
        Visitor of a name node

        :param name:
        :return: the identifier of the name
        """
        return name.id

    def visit_NameConstant(self, constant: ast.NameConstant) -> Any:
        """
        Visitor of constant names node

        :param constant:
        :return: the value of the constant
        """
        return constant.value

    def visit_Num(self, num: ast.Num) -> int:
        """
        Visitor of literal number node

        :param num:
        :return: the value of the number
        """
        return num.n

    def visit_Str(self, str: ast.Str) -> str:
        """
        Visitor of literal string node

        :param str:
        :return: the value of the string
        """
        return str.s

    def visit_Tuple(self, tup_node: ast.Tuple) -> Tuple[Any, ...]:
        """
        Visitor of literal tuple node

        :param tup_node: the python ast string node
        :return: the value of the tuple
        """
        return tuple([self.get_type(value) for value in tup_node.elts])

    def visit_List(self, list_node: ast.List) -> List[Any]:
        """
        Visitor of literal list node

        :param list_node: the python ast list node
        :return: the value of the list
        """
        return [self.get_type(value) for value in list_node.elts]
